import os
import re
import subprocess


class ShellCheck:
    def __init__(self):
        self.tmp_file = "tmp.sh"
        self.sc_rule_pattern = "SC[0-9]{4}"
        self.sc_severity_pattern = "(info)|(error)|(warning)|(style)"
        self.shellcheck = self._verify_shellcheck()

    @staticmethod
    def _verify_shellcheck():
        cmd = ["shellcheck", "-V"]
        try:
            subprocess.run(cmd, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
            return True
        except FileNotFoundError:
            print("WARNING: ShellCheck is not installed, unable to parse shell commands. "
                  "Run `pip install shellcheck-py`")
            return False

    def _write_to_file(self, command):
        with open(self.tmp_file, 'w') as f:
            f.write(command)

    @staticmethod
    def get_index(li: list, index: [int, str], offset: int = 0):
        if isinstance(index, int):
            try:
                index = (index + offset) if index >= 0 else 0
                return li[index]
            except IndexError:
                return None
        elif isinstance(index, str):
            try:
                index = (li.index(index) + offset) if li.index(index) > 0 else 0
                return li[index]
            except ValueError:
                return None

    def check(self, shell_command: str, shell: str = "sh"):
        output = []
        if self.shellcheck is False:
            return output
        self._write_to_file(command=shell_command)

        cmd = ["shellcheck", "-s", shell, self.tmp_file]
        try:
            raw_output = [i for i in subprocess.run(cmd, stderr=subprocess.STDOUT,
                                                    stdout=subprocess.PIPE).stdout.decode("utf-8").splitlines()
                          if i != ""]
        except Exception as E:
            print(f"Error: {E}")
            raw_output = []
        finally:
            os.remove(self.tmp_file)
        if len(raw_output) == 0:
            return output
        raw_output = raw_output[:raw_output.index("For more information:")]
        blobs = [raw_output.index(i) for i in raw_output if i.startswith(f"In {self.tmp_file} line")]
        for i, x in enumerate(blobs):
            start_index = x
            end_index = self.get_index(li=blobs, index=i+1)
            y = raw_output[start_index:end_index]
            output.append(
                {
                    "line_number": y[0].split(" line ", 1)[1],
                    "severity": re.search(pattern=self.sc_severity_pattern, string=y[2]).group(),
                    "sc_rule": re.search(pattern=self.sc_rule_pattern, string=y[2]).group(),
                    "sc_rule_desc": y[2].split(": ", 1)[1],
                    "wrong_line": y[1],
                    "fixed_line": self.get_index(li=y, index="Did you mean: ", offset=1)
                }
            )
        return output



import argparse
import ast
import fnmatch
import glob
import os
import re
from typing import Dict, List


class ParsingError(Exception):
    """
    Unable to parse this Dockerfile, are you sure it is valid syntax?
    """
    def __str__(self):
        return self.__doc__


class DockerfileParser:
    def __init__(self, dockerfile: str = None, raw_text: str = None, dockerignore: str = ".dockerignore"):
        self.valid_instructions = ["#", "COMMENT", "FROM", "COPY", "ADD", "WORKDIR", "EXPOSE", "USER", "ARG", "ENV",
                                   "LABEL", "RUN", "CMD", "ENTRYPOINT", "ONBUILD", "HEALTHCHECK", "STOPSIGNAL",
                                   "VOLUME", "SHELL", "MAINTAINER"]
        if dockerfile is not None:
            self.df_content = self._read_file(path=dockerfile)
        elif raw_text is not None:
            self.df_content = raw_text.splitlines()
        else:
            raise TypeError("Neither a Dockerfile path nor raw text input were provided")

        self.docker_ignore_files = self._read_file(path=dockerignore, docker_ignore=True)
        self.df_ast = self.parse_dockerfile()

    @staticmethod
    def _read_file(path: str, docker_ignore: bool = False) -> List:
        if os.path.exists(path):
            with open(path, 'r') as f:
                data = f.read().splitlines()
            return data
        else:
            if docker_ignore:
                return []
            raise FileNotFoundError(f"Dockerfile not found, path: {path}")

    def _get_state(self, line: str) -> str:
        line = line.rstrip()
        if line == "":
            state = "blank"
        elif line.strip().startswith("#"):
            state = "comment"
        elif line.split(" ", 1)[0] in self.valid_instructions and line.endswith("\\"):
            state = "new_multi_line_command"
        elif line.endswith("\\"):
            state = "continued_multi_line_command"
        elif not line.endswith("\\") and line.split(" ", 1)[0] not in self.valid_instructions:
            state = "end_multi_line_command"
        else:
            state = "new_command"
        return state

    @staticmethod
    def _get_instruction(line: str) -> [str, None]:
        if line != "":
            if line.strip().startswith("#"):
                instruction = "COMMENT"
            else:
                instruction = line.split(" ", 1)[0]
            return instruction
        return None

    def _get_raw_command(self, line: str) -> [str, None]:
        if line.startswith("#"):
            return line.split("#", 1)[1].replace("#", "")
        elif line == "":
            return None
        elif line.split(" ", 1)[0] in self.valid_instructions:
            return line.split(" ", 1)[1].strip()
        else:
            return line.strip()

    @staticmethod
    def _parse_json_notation(command: str) -> dict:
        out = dict(executable=None, arguments=[])
        try:
            run_eval = ast.literal_eval(command)
        except (SyntaxError, ValueError):
            run_eval = command
        if isinstance(run_eval, list):
            if len(run_eval) == 0:
                out["arguments"] = run_eval[1:]
                return out
            if run_eval[0].startswith("--"):
                out["docker_syntax"] = run_eval[0]
                run_eval = run_eval[1:]
            out["executable"] = run_eval[0]
            out["arguments"] = run_eval[1:]
        else:
            run_split: List = command.split(" ", 1)
            if run_split[0].startswith("--"):
                out["docker_syntax"] = run_split[0]
                run_split: List = run_split[1].split(" ", 1)
            if len(run_split) == 1:
                out["executable"] = run_split[0]
            else:
                out["executable"] = run_split[0]
                out["arguments"] = run_split[1]

        return out

    @staticmethod
    def _parse_multi_var(command, key_name):
        if "=" in command:
            env_split = command.split("=")
            return {key_name: env_split[0], 'default_value': env_split[1].strip("'").strip('"')}
        elif " " in command:
            env_split = command.split(" ", 1)
            return {key_name: env_split[0], 'default_value': env_split[1].strip("'").strip('"')}
        else:
            return {key_name: command}

    def _parse_dynamic_files(self, source_locations: list) -> List:
        to_copy_files = []
        for source_location in source_locations:
            if os.path.isfile(path=source_location):
                to_copy_files.extend([source_location])
                continue
            elif source_location.startswith("http"):
                continue
            ignored_files = []
            copy_files = glob.glob(f"./{source_location}/**", recursive=True)

            for pattern in self.docker_ignore_files:
                ignored_files.extend(fnmatch.filter(names=copy_files, pat=f"*{pattern}*"))
            to_copy_files.extend([i.split("./", 1)[1] for i in copy_files if i not in ignored_files and
                                  not i.endswith(f"{source_location}/")])
        return sorted(to_copy_files)

    def _parse_command(self, instruction: str, command: str) -> [Dict, List]:
        instruction = instruction.upper()
        if instruction == "COMMENT":
            comment = command.strip()
            if comment.startswith("#"):
                comment = comment.replace("#", "")
            return dict(comment=comment.strip())
        elif instruction.rstrip() == "":
            pass
        elif instruction.startswith("{%"):
            pass
        elif instruction == "FROM":
            image = dict(image=command)
            if " as " in command.lower():
                split = image["image"].lower().split(" as ")
                image["image"] = split[0]
                image["alias"] = split[1]
            if ":" in image["image"]:
                split = image["image"].split(":")
                image["image"] = split[0]
                image["version"] = split[1]
            return image
        elif instruction in ["COPY", "ADD"]:
            copy_split = command.split(" ")
            copy_dict = {}
            if copy_split[0].startswith("--chown"):
                copy_dict["chown"] = copy_split[0].split("=", 1)[1]
                copy_split = copy_split[1:]
            if len(copy_split) > 2:
                copy_dict["source"] = copy_split[:-1]
                copy_dict["parsed_source_files"] = self._parse_dynamic_files(source_locations=copy_split[:-1])
                copy_dict["target"] = copy_split[-1]
            else:
                copy_dict["source"] = [copy_split[0]]
                copy_dict["parsed_source_files"] = self._parse_dynamic_files(source_locations=[copy_split[0]])
                copy_dict["target"] = copy_split[1]
            return copy_dict
        elif instruction == "USER":
            if ":" in command:
                user_split = command.split(":")
                return dict(user=user_split[0], group=user_split[1])
            else:
                return dict(user=command)
        elif instruction == "ARG":
            if "=" in command:
                arg_split = command.split("=")
                return dict(argument=arg_split[0], default_value=arg_split[1])
            else:
                return dict(argument=command)
        elif instruction in ["ENV", "LABEL"]:
            envs = []
            if sum(i == "=" for i in command) > 1:
                for i in command.split(" "):
                    envs.append(self._parse_multi_var(command=i, key_name=instruction.lower()))
            else:
                envs.append(self._parse_multi_var(command=command, key_name=instruction.lower()))
            return envs
        elif instruction in ["RUN", "ENTRYPOINT", "CMD", "SHELL"]:
            return self._parse_json_notation(command=command)
        elif instruction == "EXPOSE":
            if "/" in command:
                expose_split = command.split("/")
                return dict(port=expose_split[0], protocol=expose_split[1])
            return dict(port=command)
        elif instruction in ["WORKDIR", "STOPSIGNAL", "VOLUME"]:
            return {instruction.lower(): command}
        elif instruction == "ONBUILD":
            command_split = command.replace(instruction, "").strip()
            command_split = self._parse_json_notation(command=command_split)
            return {
                "sub_instruction": command_split['executable'],
                **self._parse_json_notation(command=command_split['arguments'])
            }
        elif instruction == "HEALTHCHECK":
            pattern = re.compile("--[a-zA-Z=-]+[a-z,0-9]+")
            options = {i.replace("--", "").split("=")[0]: i.split("=", 1)[1] for i in
                       re.findall(pattern=pattern, string=command)}
            command_split = re.sub(pattern=pattern, repl="", string=command.replace(instruction, "").strip())
            command_split = self._parse_json_notation(command=command_split.strip())
            return {
                "sub_instruction": command_split['executable'],
                "options": options,
                **self._parse_json_notation(command=command_split['arguments'])
            }
        elif instruction == "MAINTAINER":
            return dict(maintainer=command)
        else:
            raise ParsingError()

    @staticmethod
    def _concat_multi_line_instruction(lines: list) -> str:
        single_line = ""
        for i in lines:
            single_line += i["raw_command"]

        lines[0]["state"] = "multi_line_command"
        lines[0]["line_number"]["end"] = max([i["line_number"]["start"] for i in lines])
        lines[0]["raw_command"] = single_line
        return lines[0]

    def split_multi_lines(self, lines: list) -> List:
        result = []
        start_index = 0
        for i, x in enumerate(lines):
            if x["state"] == "end_multi_line_command":
                result.append(self._concat_multi_line_instruction(lines=lines[start_index:i + 1]))
                start_index = i+1
        return result

    @staticmethod
    def format_and_correct_sh(instruction, raw_command: str, raw_line):
        if instruction in ["RUN", "LABEL", "ENV"]:
            command = "{} {}\n".format(instruction, raw_command.replace('\\', '\\\n\t'))
        else:
            command = f"{raw_line}\n"

        return command

    @staticmethod
    def _get_index(li, index, offset, error_value=None):
        try:
            index = index + offset
            return li[index]
        except IndexError:
            return error_value

    def parse_dockerfile(self) -> List[dict]:
        parsed = [{"line_number": dict(start=line_number + 1, end=line_number + 1),
                   "raw_line": i,
                   "state": self._get_state(line=i),
                   "instruction": self._get_instruction(line=i.strip()),
                   "raw_command": self._get_raw_command(line=i)
                   }
                  for line_number, i in enumerate(self.df_content)]

        for i, instruction in enumerate(parsed):
            next_instruction = self._get_index(li=parsed, index=i, offset=1, error_value={})
            prev_instruction = self._get_index(li=parsed, index=i, offset=-1, error_value={})
            if instruction["state"] == "comment":
                if "continued_" in next_instruction.get("state", "") or \
                        "continued_" in prev_instruction.get("state", ""):
                    parsed[parsed.index(instruction)]["state"] = ""

        multi_line_instructions = self.split_multi_lines([i for i in parsed if "_multi_line_" in i["state"]])
        single_line_instructions = sorted(
            [i for i in parsed if i["state"] in ["new_command", "comment"]] + multi_line_instructions,
            key=lambda d: d['line_number']['start']
        )
        enriched = [dict(
                         line_number=i["line_number"],
                         instruction=i["instruction"],
                         instruction_details=self._parse_command(instruction=i["instruction"],
                                                                 command=i["raw_command"]),
                         _raw=i["raw_command"],
                         formatted=self.format_and_correct_sh(instruction=i['instruction'],
                                                              raw_command=i['raw_command'], raw_line=i["raw_line"])
                         )
                    for i in single_line_instructions]
        return enriched

    def _get_instructions(self, instruction: str) -> List[dict]:
        return [i for i in self.df_ast if i["instruction"] == instruction]

    @property
    def users(self) -> List[dict]:
        return self._get_instructions(instruction="USER")

    @property
    def froms(self) -> List[dict]:
        return self._get_instructions(instruction="FROM")

    @property
    def instructions(self) -> List[dict]:
        return [i["instruction"] for i in self.df_ast]

    @property
    def copies(self) -> List[dict]:
        return self._get_instructions(instruction="COPY")

    @property
    def adds(self) -> List[dict]:
        return self._get_instructions(instruction="ADD")

    @property
    def comments(self) -> List[dict]:
        return self._get_instructions(instruction="COMMENT")

    @property
    def args(self) -> List[dict]:
        return self._get_instructions(instruction="ARG")

    @property
    def envs(self) -> List[dict]:
        return self._get_instructions(instruction="ENV")

    @property
    def labels(self) -> List[dict]:
        return self._get_instructions(instruction="LABEL")

    @property
    def runs(self) -> List[dict]:
        return self._get_instructions(instruction="RUN")

    @property
    def entrypoints(self) -> List[dict]:
        return self._get_instructions(instruction="ENTRYPOINT")

    @property
    def cmds(self) -> List[dict]:
        return self._get_instructions(instruction="CMD")

    @property
    def shells(self) -> List[dict]:
        return self._get_instructions(instruction="SHELL")

    @property
    def exposes(self) -> List[dict]:
        return self._get_instructions(instruction="EXPOSE")

    @property
    def workdirs(self) -> List[dict]:
        return self._get_instructions(instruction="WORKDIR")

    @property
    def stopsignals(self) -> List[dict]:
        return self._get_instructions(instruction="STOPSIGNAL")

    @property
    def volumes(self) -> List[dict]:
        return self._get_instructions(instruction="VOLUME")

    @property
    def healthchecks(self) -> List[dict]:
        return self._get_instructions(instruction="HEALTHCHECK")

    @property
    def onbuilds(self) -> List[dict]:
        return self._get_instructions(instruction="ONBUILD")

    @property
    def maintainers(self) -> List[dict]:
        return self._get_instructions(instruction="MAINTAINER")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--dockerfile", dest="dockerfile", required=False, help="Path to Dockerfile location")
    args = parser.parse_args()
    parser = DockerfileParser(dockerfile=args.dockerfile)
    print(parser.df_ast)



import datetime
import inspect
import json
import os
import time
import uuid

from . import __version__
from .parser import DockerfileParser
from .shellcheck import ShellCheck


class Analyzer:
    def __init__(self, dockerfile: str = None, raw_text: str = None, dockerignore: str = ".dockerignore",
                 verbose: bool = False, explain_rule: str = None, gitlab_codequality: bool = False,
                 gitlab_sast: bool = False, write_df: bool = False, show_df: bool = False,
                 silent: bool = False, **kwargs):
        if dockerfile is not None:
            self.dfp = DockerfileParser(dockerfile=dockerfile, dockerignore=dockerignore)
        elif raw_text is not None:
            self.dfp = DockerfileParser(raw_text=raw_text, dockerignore=dockerignore)
        elif explain_rule is not None:
            self.explain(rule=explain_rule)
        else:
            print("Neither a Dockerfile path nor raw text input were provided")
            exit(1)
        self.dockerfile = dockerfile
        self.raw_text = raw_text
        self.results = []
        self.results_code_climate = []
        self.shellcheck_severity_cc_map = dict(ERROR="blocker", WARNING="major", INFO="minor", STYLE="info")
        self.report = dict(INFO=0, MINOR=0, MAJOR=0, CRITICAL=0, BLOCKER=0)

        self.raw_text = True if raw_text else False
        self.verbose_explanation = verbose
        self.silent = silent
        self.gitlab_codequality = gitlab_codequality
        self.show_dockerfile = show_df
        self.write_dockerfile = write_df
        self.shellcheck = ShellCheck()
        self.kwargs = kwargs
        self.gitlab_sast = gitlab_sast
        self.gitlab_sast_severity_map = {
            "INFO": "Info",
            "MINOR": "Low",
            "MAJOR": "High",
            "CRITICAL": "Critical",
            "BLOCKER": "Critical"
        }
        self.gitlab_security_scanner = {
            "version": "14.0.4",
            "vulnerabilities": [],
            "scan": {
                "analyzer":
                    {
                        "id": "dokter",
                        "name": "Dokter",
                        "url": "https://gitlab.com/gitlab-org/incubation-engineering/ai-assist/dokter",
                        "vendor":
                            {
                                "name": "GitLab"
                            },
                        "version": "0.0.0" if __version__ == "dev" else __version__
                    },
                "scanner":
                    {
                        "id": "dokter",
                        "name": "Dokter",
                        "url": "https://gitlab.com/gitlab-org/incubation-engineering/ai-assist/dokter",
                        "vendor":
                            {
                                "name": "GitLab"
                            },
                        "version": "0.0.0" if __version__ == "dev" else __version__
                    },
                "type": "sast",
                "start_time": datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S"),
                "end_time": datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S"),
                "status": "success"
            }
        }

    def _formatter(self, rule: str, data: dict, severity: str, rule_info: str, categories: list = None):
        cc_entry = {
            "location": {
                "lines": {
                    "begin": data["line_number"]["start"],
                    "end": data["line_number"]["end"]
                },
                "path": self.dockerfile
            },
            "severity": severity,
            "type": "issue",
            "categories": categories,
            "check_name": rule.upper(),
            "description": rule_info.splitlines()[0]
        }
        self.results_code_climate.append(cc_entry)

        gss_entry = {
            "id": str(uuid.uuid4()),
            "category": "sast",
            "message": f'{rule_info.splitlines()[0]}',
            "description": rule_info.split(":return:", 1)[0],
            "cve": "",
            "severity": self.gitlab_sast_severity_map.get(severity.upper(), "Unknown"),
            "scanner": dict(id="dokter", name="Dokter"),
            "location":
                {
                    "file": self.dockerfile,
                    "start_line": data["line_number"]["start"],
                    "end_line": data["line_number"]["end"]
                },
            "identifiers": [dict(type="dokter_id", name=rule.upper(), value=str(uuid.uuid4()),
                                 url=f"https://gitlab.com/gitlab-org/incubation-engineering/ai-assist"
                                     f"/dokter/-/blob/main/docs/{rule.lower()}.md")]
        }
        self.gitlab_security_scanner["vulnerabilities"].append(gss_entry)

        if self.verbose_explanation is True:
            rule_info = f"\n{rule_info.split(':return:', 1)[0]}"
        else:
            rule_info = rule_info.splitlines()[0]

        self.results.append(f"{self.dockerfile}:{data['line_number']['start']:<3} - {rule.upper()} "
                            f"- {severity.upper():<7} - {rule_info}")

        self.report[severity.upper()] += 1

    def _return_results(self) -> dict:
        return self.report

    def dfa000(self):
        """
        Violation of Shellcheck rule

        Autocorrect: True
        :return:
        """
        categories = ["Style"]
        for i in self.dfp.runs:
            sc_results = self.shellcheck.check(
                shell_command=f'{i["instruction_details"]["executable"]} {i["instruction_details"]["arguments"]}'
            )
            for result in sc_results:
                rule = result["sc_rule"]
                if result["fixed_line"] is not None:
                    corrected = i["_raw"].replace(result["wrong_line"], result["fixed_line"])
                    i["formatted"] = self.dfp.format_and_correct_sh(instruction=i["instruction"], raw_command=corrected,
                                                                    raw_line=i["_raw"])

                severity = self.shellcheck_severity_cc_map.get(result["severity"].upper(), "info")
                self._formatter(rule=rule, severity=severity, data=i, rule_info=f'Shellcheck: {result["sc_rule_desc"]}',
                                categories=categories)

    def dfa001(self):
        """
        Verify that no credentials are leaking by copying in sensitive files.

        Examples include: copying over a .env file, SSH private keys, settings files etc.

        Autocorrect: False
        :return:
        """
        rule = inspect.stack()[0][3]
        severity = "critical"
        categories = ["Security"]
        sensitive_files = [
            ".env", ".pem", ".properties"
                            "settings", "config", "secrets", "application", "dev", "appsettings", "credentials",
            "default", "strings",
            "environment"
        ]

        for word in sensitive_files:
            for i in self.dfp.copies:
                for source in i["instruction_details"]["source"]:
                    if word in source.lower() or word in i["instruction_details"]["target"].lower():
                        self._formatter(rule=rule, data=i, severity=severity, categories=categories,
                                        rule_info=inspect.getdoc(self.dfa001))

    def dfa002(self):
        """
        Use a .dockerignore file to exclude files being copied over.

        By using a .dockerignore files, the build will generally be faster because it has to transfer less data to the
        daemon, it also helps prevent copying sensitive files. For more information see:
        https://docs.docker.com/engine/reference/builder/#dockerignore-file

        Autocorrect: False
        :return:
        """
        rule = inspect.stack()[0][3]
        severity = "info"
        categories = ["Security"]
        if len(self.dfp.docker_ignore_files) == 0:
            data = {"line_number": {"start": 0, "end": 0}}
            self._formatter(data=data, rule=rule, severity=severity, rule_info=inspect.getdoc(self.dfa002),
                            categories=categories)

    def dfa003(self):
        """
        When using "COPY . <target>" make sure to have a .dockerignore file. Best to copy specific folders.

        By using a .dockerignore files, the build will generally be faster because it has to transfer less data to the
        daemon, it also helps prevent copying sensitive files. For more information see:
        https://docs.docker.com/engine/reference/builder/#dockerignore-file

        Example of secure instruction:

        ```
        COPY src /app/src
        COPY requirements.txt /app/
        ```

        Example of insecure instruction:
        ```
        COPY . /app
        ```

        Autocorrect: False
        :return:
        """
        rule = inspect.stack()[0][3]
        severity = "major"
        categories = ["Security"]
        for i in self.dfp.copies:
            for source in i["instruction_details"]["source"]:
                if source == ".":
                    self._formatter(rule=rule, data=i, severity=severity, rule_info=inspect.getdoc(self.dfa003),
                                    categories=categories)

    def dfa004(self):
        """
        Verify that build args doesn't contain sensitive information, use secret mounts instead.

        Build args are stored in the history of the docker image and can be retrieved. Secret mounts are not persisted
        and are therefor a better option if you temporarily need sensitive information to build your image. If sensitive
        information is required during runtime of the containers, use environment variables.

        Example of secure instruction:

        ```
        RUN --mount=type=secret,id=docker_token docker login -u user -p $(cat /run/secrets/docker_token)
        ```

        Example of insecure instruction:
        ```
        ARG TOKEN
        RUN docker login -u user -p $TOKEN
        ```

        Autocorrect: False
        :return:
        """
        rule = inspect.stack()[0][3]
        severity = "critical"
        categories = ["Security"]

        sensitive_words = [
            "key", "secret", "token", "pass"
        ]

        for word in sensitive_words:
            for i in self.dfp.args:
                if word.lower() in i["instruction_details"]["argument"].lower():
                    self._formatter(data=i, severity=severity, rule=rule, rule_info=inspect.getdoc(self.dfa004),
                                    categories=categories)

    def dfa005(self):
        """
        Don't use root but use the least privileged user.

        In a Docker container a root user is the same UID as root on the machine, this could be exploited. After doing
        things required by root always switch back to the least privileged user.

        Example of secure instruction:

        ```
        FROM python:3.10.0
        RUN useradd -D appuser && chown -R appuser /app
        USER appuser
        CMD ["python", "main.py"]
        ```

        Example of insecure instruction:
        ```
        FROM python:3.10.0
        CMD ["python", "main.py"]
        ```

        Autocorrect: True
        :return:
        """
        rule = inspect.stack()[0][3]
        severity = "major"
        categories = ["Security"]
        if len(self.dfp.users) > 0:
            last_user = self.dfp.users[-1]
            if len(self.dfp.users) > 0 and last_user["instruction_details"]["user"].lower() == "root":
                self._formatter(data=last_user, severity=severity, rule=rule, rule_info=inspect.getdoc(self.dfa005),
                                categories=categories)
                workdir = "/app" if len(self.dfp.workdirs) == 0 else \
                    self.dfp.workdirs[-1]["instruction_details"]["workdir"]
                last_user["formatted"] = f"WORKDIR {workdir}\nRUN useradd -M appuser &&" \
                                         f" chown -R appuser {workdir}\nUSER appuser\n"

    def dfa006(self):
        """
        The name of the Dockerfile must be 'Dockerfile' or a pattern of '<purpose>.Dockerfile'

        To ensure contents are recognized as a Dockerfile and correctly rendered in IDE's and version control systems.

        Good:
        - Dockerfile
        - api.Dockerfile
        - dev.Dockerfile
        - api.dev.Dockerfile

        Neutral:
        - Dockerfile.api

        Bad:
        - dockerfile
        - DockerFile
        - Dockerfile1
        - Dockerfile-api

        Autocorrect: False
        :return:
        """
        rule = inspect.stack()[0][3]
        severity = "minor"
        categories = ["Style"]
        if os.path.basename(self.dockerfile).split(".")[-1] != "Dockerfile":
            data = {"line_number": {"start": 0, "end": 0}}
            self._formatter(rule=rule, data=data, severity=severity, rule_info=inspect.getdoc(self.dfa006),
                            categories=categories)

    def dfa007(self):
        """
        Only use ADD for downloading from a URL or automatically unzipping local files, use COPY for other local files.

        See also: https://docs.docker.com/develop/develop-images/dockerfile_best-practices/#add-or-copy

        Autocorrect: True
        :return:
        """
        rule = inspect.stack()[0][3]
        severity = "minor"
        categories = ["Bug Risk"]
        for i in self.dfp.adds:
            for source in i["instruction_details"]["source"]:
                # Docker actually checks if a file is compressed regardless of name, but this is a good first step
                if source.startswith("http") is True:
                    pass
                elif source.endswith(".gz") is True:
                    pass
                else:
                    self._formatter(rule=rule, data=i, severity=severity, rule_info=inspect.getdoc(self.dfa007),
                                    categories=categories)
                    i["formatted"] = i["formatted"].replace("ADD ", "COPY ")

    def dfa008(self):
        """
        Chain multiple RUN instructions together to reduce the number of layers and size of the image.

        Autocorrect: True
        :return:
        """
        rule = inspect.stack()[0][3]
        severity = "major"
        categories = ["Performance"]
        first_run = None
        for i, instruction in enumerate(self.dfp.instructions):
            if instruction == "RUN" and instruction == self.dfp.df_ast[i - 1]["instruction"]:
                self._formatter(rule=rule, severity=severity, data=self.dfp.df_ast[i], categories=categories,
                                rule_info=inspect.getdoc(self.dfa008))
                if first_run is None:
                    first_run = self.dfp.df_ast[i - 1]
                corrected = self.dfp.df_ast[i]['formatted'].replace(f"{self.dfp.df_ast[i]['instruction']} ", '')
                first_run["formatted"] = first_run["formatted"][:-1] + " && \\\n" + "\t" + corrected
                # Delete the unneeded RUN statement
                del self.dfp.df_ast[i]["formatted"]

    def dfa009(self):
        """
        To be added: Follow correct order to optimize caching.
        :return:
        """
        pass

    def dfa010(self):
        """
        Include a healthcheck for long-running or persistent containers.

        Autocorrect: False
        :return:
        """
        rule = inspect.stack()[0][3]
        severity = "info"
        categories = ["Performance"]
        if "HEALTHCHECK" not in self.dfp.instructions:
            data = {"line_number": {"start": 0, "end": 0}}
            self._formatter(rule=rule, data=data, severity=severity, rule_info=inspect.getdoc(self.dfa010),
                            categories=categories)

    def dfa011(self):
        """
        CMD or ENTRYPOINT should be the last instruction.
        :return:
        """
        rule = inspect.stack()[0][3]
        severity = "major"
        categories = ["Style"]

        # Check if multi-stage
        df_ast = self.dfp.df_ast
        dfp_instructions = self.dfp.instructions
        froms = [(n, i) for n, i in enumerate(self.dfp.instructions) if i == "FROM"]
        if len(froms) > 1:
            multi_stage_offset = froms[-1][0]
            dfp_instructions = self.dfp.instructions[multi_stage_offset:]
            df_ast = self.dfp.df_ast[multi_stage_offset:]

        instructions_past_entrypoint = []
        if "ENTRYPOINT" in dfp_instructions:
            instructions_past_entrypoint = df_ast[dfp_instructions.index("ENTRYPOINT") + 1:]

        instructions_past_cmd = []
        if "CMD" in dfp_instructions:
            instructions_past_cmd = df_ast[dfp_instructions.index("CMD") + 1:]

        for i in instructions_past_entrypoint + instructions_past_cmd:
            if i["instruction"] not in ["CMD", "COMMENT"]:
                self._formatter(rule=rule, severity=severity, data=i, rule_info=inspect.getdoc(self.dfa011),
                                categories=categories)

    def dfa012(self):
        """
        MAINTAINER is deprecated, use LABEL instead.

        Incorrect:
        ```
        MAINTAINER dev@someproject.org
        ```

        Correct:
        ```
        LABEL maintainer="dev@someproject.org"
        ```

        Autocorrect: True
        :return:
        """
        rule = inspect.stack()[0][3]
        severity = "major"
        categories = ["Style"]
        if len(self.dfp.maintainers) > 0:
            for i in self.dfp.maintainers:
                self._formatter(rule=rule, severity=severity, data=i, rule_info=inspect.getdoc(self.dfa012),
                                categories=categories)
                i["formatted"] = i["formatted"].replace("MAINTAINER ", "LABEL maintainer=")

    @staticmethod
    def _write_file(location, data):
        with open(location, "w") as f:
            f.write(data)

    def formatter(self):
        a = [i for i in self.dfp.df_ast if i.get("formatted") is not None]
        data = ""
        for line, instruction in enumerate(a):
            next_instruction = self.shellcheck.get_index(li=a, index=line, offset=1)
            curr_instruction = instruction["instruction"]
            if curr_instruction == "COMMENT":
                data += instruction['formatted']
            elif next_instruction is not None:
                if curr_instruction != next_instruction["instruction"]:
                    data += f"{instruction['formatted']}\n"
                else:
                    data += instruction['formatted']
            else:
                data += f"{instruction['formatted']}\n"
        return data

    def run(self):
        # A bit of a hack to run all the rules
        for f in [i for i, f in inspect.getmembers(object=Analyzer) if i.startswith("dfa")]:
            getattr(Analyzer, f)(self)

        # Print the results
        if self.silent is False:
            for i in sorted(set(self.results)):
                print(i)

        if self.gitlab_codequality:
            report_location = f"dokter-{os.environ.get('CI_COMMIT_SHA', int(time.time()))}.json"
            self._write_file(location=report_location, data=json.dumps(self.results_code_climate))
            print(f"\nCode Quality report written to: {report_location}")

        if self.gitlab_sast:
            self.gitlab_security_scanner["end_time"] = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
            report_location = f"dokter-sast-{os.environ.get('CI_COMMIT_SHA', int(time.time()))}.json"
            self._write_file(location=report_location, data=json.dumps(self.gitlab_security_scanner))
            print(f"\nSAST report written to: {report_location}")

        new_dockerfile = self.formatter()

        if self.write_dockerfile:
            report_location = "Dockerfile"
            if self.dockerfile is not None:
                report_location = self.dockerfile
            self._write_file(location=report_location, data=new_dockerfile)
            print(f"\nNew Dockerfile written to: {report_location}")

        if self.show_dockerfile:
            print("This is the new Dockerfile:\n")
            print(new_dockerfile)

        return self.report

    @staticmethod
    def explain(rule):
        try:
            out = f"Explanation for rule '{rule}':\n{getattr(Analyzer, rule.lower()).__doc__.split(':return:', 1)[0]}"
        except AttributeError:
            out = "Rule does not exists"
        print(out)
        return out



import argparse

from . import __version__
from .analyzer import Analyzer


def dokter():
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--dockerfile", dest="dockerfile", required=False, help="Path to Dockerfile location")
    parser.add_argument("-e", "--explain", dest="explain_rule", required=False, help="Explain what a rule entails")
    parser.add_argument("-c", "--gitlab-codequality", dest="gitlab_codequality", action="store_true", required=False,
                        help="Save the output in a JSON formatted for GitLab Code Quality reports")
    parser.add_argument("--sast", dest="gitlab_sast", action="store_true", required=False,
                        help="Save the output in a JSON formatted for GitLab SAST reports")
    parser.add_argument("-w", "--write-dockerfile", dest="write_df", action="store_true", required=False,
                        help="Save the output in a JSON formatted for GitLab Code Quality reports")
    parser.add_argument("-s", "--show-dockerfile", dest="show_df", action="store_true", required=False,
                        help="Save the output in a JSON formatted for GitLab Code Quality reports")
    parser.add_argument("-V", "--verbose", dest="verbose", required=False, action="store_true",
                        help="Verbose information")
    parser.add_argument("--dockerignore", dest="dockerignore", required=False, default=".dockerignore",
                        help="Path to dockerignore file, default: .dockerignore")
    parser.add_argument("-v", "--version", dest="version", required=False, action="store_true",
                        help="Print version of Dokter")
    args = parser.parse_args()
    if args.version is True:
        print(__version__)
        exit(0)

    a = Analyzer(**vars(args))
    if args.explain_rule is None:
        result = a.run()
        del result["INFO"]
        if sum(result.values()) > 0:
            exit(1)


if __name__ == "__main__":
    dokter()
